/*
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * NodeOS Channel API
 */
#ifndef _NODEOS_CHANNEL_H_
#define _NODEOS_CHANNEL_H_

#include <nodeos/resource.h>
#include <nodeos/demuxkey.h>
#include <nodeos/packet.h>

typedef struct {
	int stuff;
} *an_chan_t;

/*
 * Channel peer address information.
 * Channel-type specific.
 */
typedef struct {
	int len;
	char data[128-sizeof(int)];
} an_chanaddr_t;

/*
 * Channel creation information.
 * To shorten the channel creation call, we gather all the parameters up
 * in a single structure.
 */
typedef struct {
	char *name;		/* the "pathname" of the channel (all) */
	int type;		/* In, out, cut (all) */
	an_chanaddr_t addr;	/* protocol specific peer address info (all) */
	an_cpuspec_t cpu;	/* CPU (packet receipt) resources (in, cut) */
	an_memspec_t mem;	/* Memory (buffering) resources (in, cut) */
	an_netspec_t net;	/* Network (bandwidth) resources (out, cut) */
} an_chanattr_t;

/* channel types */
#define inChan	1
#define outChan	2
#define cutChan	3

/*
 * Called when a packet arrives on a channel
 */
typedef void (*an_chanrecv_t)(an_packet_t);

/* XXX compat */
#ifdef NODEOS_SPEC_COMPAT
#define an_Channel an_chan_t
#define an_ChannelAddress an_chanaddr_t
#define an_ChannelAttr an_chanattr_t
#define an_DeliverFunc an_chanrecv_t
#define an_channelCreate an_chan_create
#define an_channelDestroy an_chan_destroy
#define an_channelSend an_chan_send
#define an_channelGetAddr an_chan_getaddr
#define an_channelAddMem an_chan_addmem
#endif

/*
 * Create a channel as described by information in the attribute structure.
 * Returns a pointer to the channel object or zero on an error.
 *
 * For in-channels, packets destined for the channel are identified by
 * the channel-type specific addressing information and the demuxkey object
 * passed in.  Receipt of a "matching" packet results in invocation of the
 * provided packet receiver function.
 *
 * For out-channels, only the attribute information is valid.
 */
an_chan_t an_chan_create(an_chanattr_t *attr, an_demuxkey_t key,
			 an_chanrecv_t func);

/*
 * Cleanly shutdown a channel.
 * Further operations on the channel should return an error.  For
 * in-channels, no further packets will be received, though processing
 * of currently active packets continues.
 */
void an_chan_destroy(an_chan_t chan);

/*
 * Send the indicated packet to an out-channel.
 * Once this function returns, the packet may be safely destroyed or
 * returned to the channel buffer pool.  No indication is provided as
 * to whether the packet was successfully transmitted.  Returns an error
 * if the channel has been destroyed.
 */
an_error_t an_chan_send(an_chan_t chan, an_packet_t packet);

/*
 * Return the address information for the indicated channel.
 * XXX hack, hack: this only exists to get the HW address for an ethernet
 * in-channel.  Should be replaced by a more generic node information
 * interface.
 */
an_error_t an_chan_getaddr(an_chan_t chan, an_chanaddr_t *addr);

/*
 * Allocate additional buffer memory for a channel.
 * The indicated number of pages of memory are allocated and broken
 * into buffers for receiving packets on an in-channel.  Along with
 * an_packet_return, this is how buffer memory is replenished.
 * Returns an error if applied to a dead channel or an out-channel.
 * XXX number of pages is probably not the right unit for this.
 */
an_error_t an_chan_addmem(an_chan_t chan, int npages);

#endif /* _NODEOS_CHANNEL_H_ */
