/*
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * A "clock" channel driver.
 *
 * In-channel only.  Packets are delivered at a rate given by the address info.
 * (e.g., if address is "5", a packet is delivered every 5th clock tick).
 * The delivered packet contains the 4-byte internal time value (``mytime'').
 */
#include "options.h"

#ifdef CHAN_CLOCK

#include <errno.h>
#include <sys/time.h>
#ifdef OSKIT
#include <oskit/com/listener.h>
#include <oskit/dev/dev.h>
#include <oskit/dev/clock.h>
#include <oskit/dev/timer.h>
#else
#define oskit_iunknown_t void
#define oskit_error_t int
#endif

#include "channel.h"
#include "packet.h"
#include "thread.h"
#include "synch.h"

#define MAXCHAN	100
#define FREQ	100

#ifdef OSKIT
extern oskit_listener_t *oskit_create_threaded_listener(
			oskit_listener_callback_t handler, void *arg);
extern oskit_clock_t *oskit_system_clock;

static oskit_timer_t *chan_clock_timer;
static oskit_itimerspec_t chan_ts;
#else
static void *chan_clock_clock(void *arg);
#endif

static oskit_error_t chan_clock_intr(oskit_iunknown_t *l, void *arg);

static volatile unsigned int mytime;

static struct chans {
	struct chans *next;
	channel_t *chan;
	int freq;
} chans[MAXCHAN], *allchans, *freechans;
static ilock_t clocklock;

struct clockaddr {
	int freq;
};

void
chan_clock_init(void)
{
#ifdef OSKIT
	oskit_listener_t *l;
#else
	pthread_t id;
	int err;
#endif
	int i;

	ilock_init(&clocklock);
	ilock_lock(&clocklock);

	for (i = 0; i < MAXCHAN; i++) {
		chans[i].next = freechans;
		freechans = &chans[i];
	}

#ifdef OSKIT
	/*
	 * Create a timer with a thread-aware listener.
	 * XXX the handler runs without a flow context.
	 */
	l = oskit_create_threaded_listener(chan_clock_intr, 0);
	oskit_clock_createtimer(oskit_system_clock, &chan_clock_timer);
	oskit_timer_setlistener(chan_clock_timer, l);
	oskit_listener_release(l);
	chan_ts.it_value.tv_sec = 0;
	chan_ts.it_value.tv_nsec = 1000000000 / FREQ;
	oskit_timer_settime(chan_clock_timer, 0, &chan_ts);
#else
	/*
	 * XXX hack: create a thread to just do select-as-timer and invoke
	 * the interrupt handler whenever it wakes up.  We could do this for
	 * the OSKit as well, but select is very heavyweight in the OSKit.
	 */
	err = pthread_create(&id, 0, chan_clock_clock, 0);
	assert(err == 0);
	err = pthread_detach(id);
	assert(err == 0);
#endif

	ilock_unlock(&clocklock);
}

int
chan_clock_create(channel_t *chan)
{
	int freq;
	struct chans *ch;
	struct clockaddr *ca = (struct clockaddr *)channel_addrdata(chan);

	freq = ca->freq;
	if (freq < 1)
		return EINVAL;

	ilock_lock(&clocklock);
	ch = freechans;
	freechans = ch->next;
	ch->chan = chan;
	ch->freq = freq;
	ch->next = allchans;
	allchans = ch;
	ilock_unlock(&clocklock);

	return 0;
}

void
chan_clock_destroy(channel_t *chan)
{
	struct chans *ch, **pch;

	ilock_lock(&clocklock);
	for (pch = &allchans; (ch = *pch) != 0; pch = &ch->next)
		if (ch->chan == chan) {
			*pch = ch->next;
			ch->next = freechans;
			freechans = ch;
			break;
		}
	ilock_unlock(&clocklock);
}

static oskit_error_t
chan_clock_intr(oskit_iunknown_t *l, void *arg)
{
	struct chans *ch;
	packet_t *packet;
	unsigned int ot;
	int err;

	ilock_lock(&clocklock);
	ot = ++mytime;
	for (ch = allchans; ch != 0; ch = ch->next)
		if (ch->freq == 1 || (ot % ch->freq) == 0) {
			err = channel_packet_create(ch->chan, &ot, sizeof ot, 1,
						    &packet, 0);
			if (err == 0)
				channel_thread_start(ch->chan, packet);
		}
	ilock_unlock(&clocklock);

#ifdef OSKIT
	oskit_timer_settime(chan_clock_timer, 0, &chan_ts);
#endif
	return 0;
}

#ifndef OSKIT

static void *
chan_clock_clock(void *arg)
{
	struct timeval delta;

	delta.tv_sec = 0;

	while (1) {
		delta.tv_usec = 1000000 / FREQ;
		(void) select(0, 0, 0, 0, &delta);
		chan_clock_intr(0, 0);
	}

	return 0;
}

#endif

#endif /* CHAN_CLOCK */
