/*
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * A raw ethernet channel driver.
 */
#include "options.h"

#ifdef CHAN_ETH

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <assert.h>

#ifdef OSKIT
#include <oskit/startup.h>
#include <oskit/dev/dev.h>
#include <oskit/dev/ethernet.h>
#include <oskit/dev/net.h>
#include <oskit/net/ether.h>
#include <oskit/io/bufio.h>
#else
#include <net/if.h>
#include <netinet/if_ether.h>

#define ETHER_HDR_SIZE ETH_HLEN

#include "unix_eth.h"
#endif

#include "nodeos.h"
#include "channel.h"
#include "demux.h"
#include "demux_tlist.h"
#include "packet.h"
#include "thread.h"
#include "ilock.h"
#include "mem.h"

struct chandata;

/*
 * Per-device structure.
 * Most oskit drivers restrict to only a single open of a device,
 * we use a single send/recv netio shared by many channels.
 *
 * The etherdev lock is lower in the locking hierarchy than the channel
 * lock.
 */
struct etherdev {
	ilock_t		 lock;
	char		 name[8];	/* "eth0" to "eth9999" */
	unsigned char	 addr[6];	/* our address */
	demux_t		 *demux;	/* list of demuxers for this device */

#ifdef OSKIT        
	/* OSKit specific */
	oskit_etherdev_t *eth;		/* device handle */
	oskit_devinfo_t	 info;		/* device driver info (unused) */
	oskit_netio_t	 *send;		/* handle for outgoing packets */
	oskit_netio_t	 *recv;		/* handle for incoming packets */
#else
	char		if_name[IFNAMSIZ]; /* system ethernet device name */
	int		recv;           /* socket */
#endif        
};

/*
 * Ethernet channel private data.  Protected by the channel lock.
 * Contains a pointer to the etherdev structure and, for inChan channels,
 * an array of received packet descriptors.
 */
struct chandata {
	channel_t	 *chan;		/* channel we are a part of */
	struct etherdev	 *edev;		/* associated ether device */

	/* OSKit specific */
	struct chanpdata {
#ifdef OSKIT        
		oskit_bufio_t	*bufio;
		oskit_size_t	size;
#else
		char		*data;
		size_t		size;
#endif                
	} pdata[0];			/* received packet data */
};

/* XXX this should possibly be exposed */
struct chanethaddr {
	unsigned char addr[6];
	unsigned short type;
};


static struct etherdev *etherdevs;
static int netherdevs;
static char chanprefix[] = "eth";
static ilock_t ethlock;

static struct etherdev *chan_edev_find(char *name);
static int chan_edev_demux_add(struct etherdev *edev, channel_t *chan);
static void chan_edev_demux_remove(struct etherdev *edev, channel_t *chan);
static int chan_edev_init(struct etherdev *edev);
static void chan_eth_receive(void *arg);

#ifdef OSKIT
static oskit_error_t chan_edev_intr(void *arg, oskit_bufio_t *b,
				    oskit_size_t psize);

extern oskit_netio_t *oskit_threaded_netio_create(
  oskit_error_t (*func)(void *data, oskit_bufio_t *b, oskit_size_t pkt_size),
  void *data);
#endif

void
chan_eth_init(void)
{
	int ndev, i;
#ifdef OSKIT
	oskit_etherdev_t **edev;
#else
	struct if_info *edev;
#endif        

	ilock_init(&ethlock);
	ilock_lock(&ethlock);

	/*
	 * Look up all the ethernet devices
	 */
#ifdef OSKIT
	ndev = osenv_device_lookup(&oskit_etherdev_iid, (void***)&edev);
#else 
	ndev = unix_eth_get_if_info(&edev);
#endif
	if (ndev <= 0) {
                printf("%s: no devices!\n", __FUNCTION__);
		return;
        }

	/*
	 * Create a device to name mapping
	 */
	etherdevs = mem_alloc(ndev * sizeof(struct etherdev));
	if (etherdevs == 0)
		return;
	memset(etherdevs, 0, ndev * sizeof(struct etherdev));
	for (i = 0; i < ndev; i++) {
		int j;
#ifdef OSKIT
                int err;
#endif

		snprintf(etherdevs[i].name, sizeof etherdevs[i].name, "%s%d",
			 chanprefix, i);
#ifdef OSKIT
		etherdevs[i].eth = edev[i];

		err = oskit_etherdev_getinfo(edev[i], &etherdevs[i].info);
		assert(err == 0);
		oskit_etherdev_getaddr(edev[i], etherdevs[i].addr);
                
		printf("%s: %-16s", etherdevs[i].name, etherdevs[i].info.name);

#else 
		strncpy(etherdevs[i].if_name, edev[i].name, IFNAMSIZ);
		memcpy(etherdevs[i].addr, edev[i].addr, 6);

		printf("%s: %-16s", etherdevs[i].name, etherdevs[i].if_name);
#endif

		for (j = 0; j < 6; j++)
			printf("%c%02x", j==0? ' ' : ':', etherdevs[i].addr[j]);
		printf("\n");

		ilock_init(&etherdevs[i].lock);
	}
	netherdevs = ndev;
	free(edev);

	ilock_unlock(&ethlock);
}

/*
 * Hook up to the ethernet device.
 * Called with the channel locked.
 */
int
chan_eth_create(channel_t *chan)
{
	struct chandata *cdata;
	struct etherdev *edev;
	size_t size;
	int err;

	/* XXX not yet */
	if (channel_type(chan) == cutChan)
		return EINVAL;

	/*
	 * Lookup the device corresponding to the name given
	 */
	edev = chan_edev_find(channel_name(chan));
	if (edev == 0)
		return EINVAL;

	/*
	 * Create the ethernet specific data struct.
	 * In addition to the base info, we allocate one bufio per
	 * possible service thread.
	 */
	size = sizeof(struct chandata);
	if (channel_type(chan) == inChan)
		size += sizeof(struct chanpdata) * channel_maxthreads(chan);
	cdata = mem_alloc(size);
	if (cdata == 0)
		return ENOMEM;

	memset(cdata, 0, size);
	cdata->edev = edev;
	/* XXX init a free/inuse list of chanpdata */

	/*
	 * Now hook up to the ethernet device
	 */
	ilock_lock(&edev->lock);

	/*
	 * If device hasn't been initialized, do so now
	 */
	if (edev->recv == 0) {
		err = chan_edev_init(edev);
		if (err) {
			ilock_unlock(&edev->lock);
			mem_free(cdata, size);
			return err;
		}
	}
	assert(edev->recv);

	/*
	 * Hook up channel and local data
	 */
	chan->data = cdata;
	cdata->chan = chan;

	/*
	 * Add the demux filter based on the address and demux info.
	 */
	if (channel_type(chan) == inChan) {
		struct chanethaddr *eaddr;

		/*
		 * XXX fill in our known HW address since right now there is no
		 * way for an EE to get the HW address and provide it.
		 */
		eaddr = (struct chanethaddr *)channel_addrdata(chan);
		if (eaddr->addr[0] != 0xff) {
			memcpy(eaddr->addr, edev->addr, sizeof edev->addr);
			channel_addrlen(chan) = sizeof edev->addr;
		}

		err = chan_edev_demux_add(edev, chan);
		if (err) {
			ilock_unlock(&edev->lock);
			mem_free(cdata, size);
			return err;
		}
	}

	ilock_unlock(&edev->lock);
	return 0;
}

/*
 * Called with the channel locked
 */
void
chan_eth_destroy(channel_t *chan)
{
	struct chandata *cdata;
	struct etherdev *edev;
	size_t size;

	cdata = chan->data;
	size = sizeof(struct chandata);
	if (channel_type(chan) == inChan)
		size += sizeof(struct chanpdata) * channel_maxthreads(chan);

	edev = cdata->edev;
	ilock_lock(&edev->lock);

	/* XXX kill service threads, free up packets/bufios */
	
	/*
	 * Remove the demux
	 */
	if (chan->demux)
		chan_edev_demux_remove(edev, chan);

	ilock_unlock(&edev->lock);

	mem_free(cdata, size);
}

/*
 * Send a packet out on a channel.
 * Called with the channel locked.
 */
void
chan_eth_send(channel_t *chan, packet_t *pack)
{
#ifdef OSKIT
	oskit_bufio_t *bio;
	oskit_size_t cc, size;
	int err;
	void *data;        
	struct ether_header *eth;
#else
	size_t size;
	struct ether_header eth;
#endif
	struct etherdev *edev;
	struct chanethaddr *eaddr;
        
	/*
	 * Create a bufio for the packet and header.
	 */
	size = pack->size + ETHER_HDR_SIZE;
#ifdef OSKIT        
	bio = oskit_bufio_create(size);
	if (bio == 0)
		return;
#endif        

	assert(channel_addrlen(chan) >= sizeof *eaddr);
	eaddr = (struct chanethaddr *)channel_addrdata(chan);
	edev = ((struct chandata *)(chan->data))->edev;

#ifndef OSKIT
	memcpy(eth.ether_dhost, eaddr->addr, sizeof eaddr->addr);
	memcpy(eth.ether_shost, edev->addr, sizeof edev->addr);
	eth.ether_type = eaddr->type;
        
	os_write_packet(edev->recv, packet_data(pack), packet_size(pack), &eth);
#else        
	/*
	 * Copy header info and data into the bufio.
	 * We do this by mapping if possible, by writing otherwise.
	 */
	err = oskit_bufio_map(bio, &data, 0, size);
	if (err == 0) {
		eth = data;
		memcpy(eth->ether_dhost, eaddr->addr, sizeof eaddr->addr);
		memcpy(eth->ether_shost, edev->addr, sizeof edev->addr);
		eth->ether_type = eaddr->type;
		memcpy(data + ETHER_HDR_SIZE, packet_data(pack),
		       packet_size(pack));
		oskit_bufio_unmap(bio, data, 0, size);
	} else {
		struct ether_header hdr;

		eth = &hdr;
		memcpy(eth->ether_dhost, eaddr->addr, sizeof eaddr->addr);
		memcpy(eth->ether_shost, edev->addr, sizeof edev->addr);
		eth->ether_type = eaddr->type;
		err = oskit_bufio_write(bio, eth, 0, ETHER_HDR_SIZE, &cc);
		if (err != 0 || cc != ETHER_HDR_SIZE) {
			oskit_bufio_release(bio);
			return;
		}
		err = oskit_bufio_write(bio, packet_data(pack), ETHER_HDR_SIZE,
					packet_size(pack), &cc);
		if (err != 0 || cc != packet_size(pack)) {
			oskit_bufio_release(bio);
			return;
		}
	}

	/*
	 * Push the packet out.
	 * We lock the edev here but that is probably unnecessary as we
	 * are serialized by the channel lock.
	 */
	ilock_lock(&edev->lock);
	(void) oskit_netio_push(edev->send, bio, size);
	ilock_unlock(&edev->lock);
	oskit_bufio_release(bio);
#endif /* OSKIT */        
}

static struct etherdev *
chan_edev_find(char *name)
{
	char *cp;
	int i, len;

	cp = index(name, '/');
	if (cp++ == 0)
		return 0;
	len = strlen(cp);

	ilock_lock(&ethlock);
	for (i = 0; i < netherdevs; i++)
		if (strncmp(etherdevs[i].name, cp, len) == 0) {
			ilock_unlock(&ethlock);
			return &etherdevs[i];
		}

	ilock_unlock(&ethlock);
	return 0;
}

/*
 * Register channel demux information with the device.  We incorporate the
 * address and type information into the demux as well.
 * Called with the channel and etherdev locked.
 */
static int
chan_edev_demux_add(struct etherdev *edev, channel_t *chan)
{
	struct chanethaddr *eaddr;
	demux_tlist_t *list;
	int err;

	/*
	 * Expand the existing demux list (or create a new one)
	 */
	if (chan->demux == 0)
		err = demux_tlist_create(2, &list);
	else {
		err = demux_tlist_copy(chan->demux, chan->demux->ntriples+2,
				       &list);
		if (err == 0)
			demux_tlist_adjust(list, ETHER_HDR_SIZE);
	}
	if (err)
		return err;

	/*
	 * Tack on matches for the address and type information
	 */
	eaddr = (struct chanethaddr *)channel_addrdata(chan);
	err = demux_tlist_prepend(list, 12, 2, &eaddr->type);
	if (err == 0)
		err = demux_tlist_prepend(list, 0, 6, eaddr->addr);
	if (err) {
		demux_tlist_destroy(list);
		return err;
	}

	/*
	 * Replace the old channel demux info with our concoction
	 * and register the new info.
	 *
	 * XXX deref the original demuxkey object?
	 */
	chan->demux = list;
	err = channel_demux_add(chan, &edev->demux);
	if (err) {
		demux_tlist_destroy(list);
		chan->demux = 0;
	}

	return err;
}

static void
chan_edev_demux_remove(struct etherdev *edev, channel_t *chan)
{
	channel_demux_remove(chan, &edev->demux);
	demux_tlist_destroy(chan->demux);
	chan->demux = 0;
}

#ifdef OSKIT
/*
 * Receive netio push function, called from the interrupt handler but in
 * a full thread context.  Bufio comes in with a ref we must release to
 * free the buffer.
 *
 * This routine serializes all packets coming in on the netio (currently one
 * netio per interface) so lets get in and out as quick as possible.  Just
 * queue the bufio and fire up a flow thread.
 */
static oskit_error_t
chan_edev_intr(void *arg, oskit_bufio_t *b, oskit_size_t psize)
{
	struct etherdev *edev = arg;
	channel_t *chan;
	struct chandata *cdata;
	struct chanpdata *cdp, *ecdp;
	void *data;

	/*
	 * Map the packet for demux examination
	 */
	assert(psize >= ETHER_HDR_SIZE);
	if (oskit_bufio_map(b, &data, 0, psize) != 0) {
		oskit_bufio_release(b);
		return 0;
	}

	/*
	 * Find a channel for this packet.
	 * If no channel matches, the packet is dropped.
	 */
	ilock_lock(&edev->lock);
	chan = demux_apply(edev->demux, data, psize);
	ilock_unlock(&edev->lock);
	oskit_bufio_unmap(b, data, 0, psize);
	if (chan == 0) {
		oskit_bufio_release(b);
		return 0;
	}

	channel_lock(chan);

	/* XXX channel shutdown has begun */
	if (chan->flags & CHAN_DYING) {
		oskit_bufio_release(b);
		channel_unlock(chan);
		return 0;
	}

	/*
	 * Find an empty channel packet data entry.
	 * XXX this could be improved obviously
	 */
	cdata = chan->data;
	ecdp = &cdata->pdata[channel_maxthreads(chan)];
	for (cdp = cdata->pdata; cdp != ecdp; cdp++)
		if (cdp->bufio == 0)
			break;

	/*
	 * No bufios available or a failure to start a thread indicate
	 * no CPU resources.  Just drop the packet.
	 */
	if (cdp == ecdp ||
	    thread_start_kernel(chan->tpool, chan, chan_eth_receive, chan, 0)) {
		channel_unlock(chan);
		oskit_bufio_release(b);
		return EAGAIN; /* XXX */
	}

	/*
	 * Queue the bufio/size and get out.  Releasing the lock will allow
	 * the flow thread to do its work.
	 */
	cdp->bufio = b;
	cdp->size = psize;
	channel_unlock(chan);

	return 0;
}

#else

void *
chan_edev_read_loop(void *arg)
{
        struct etherdev *edev = arg;
        int count;
        char *buf;
        int bufsize;
	channel_t *chan;
	struct chandata *cdata;
	struct chanpdata *cdp, *ecdp;

	bufsize = NODEOS_CHAN_BUFSIZE;

	while (1) {
		buf = mem_alloc(bufsize);
		assert(buf != 0);

	again:
		count = os_read_packet(edev->recv, buf, bufsize);

		/*
		 * Find a channel for this packet.
		 * If no channel matches, the packet is dropped.
		 */
		ilock_lock(&edev->lock);
		chan = demux_apply(edev->demux, buf, count);
		ilock_unlock(&edev->lock);
		if (chan == 0)
			goto again;

		channel_lock(chan);

		/* XXX channel shutdown has begun */
		if (chan->flags & CHAN_DYING) {
			channel_unlock(chan);
			break;
		}

		/*
		 * Find an empty channel packet data entry.
		 * XXX this could be improved obviously
		 */
		cdata = chan->data;
		ecdp = &cdata->pdata[channel_maxthreads(chan)];
		for (cdp = cdata->pdata; cdp != ecdp; cdp++)
			if (cdp->data == 0)
				break;

		/*
		 * No available buffer slot or a failure to start a thread
		 * indicates no CPU resources.  Just drop the packet.
		 */
		if (cdp == ecdp ||
		    thread_start_kernel(chan->tpool, chan, chan_eth_receive,
					chan, 0)) {
			channel_unlock(chan);
			goto again;
		}

		/*
		 * Otherwise, queue the buffer and get back to work.
		 * Releasing the lock will allow the flow thread to do
		 * its work.  The flow thread will free the buffer when
		 * it is done.
		 */
		cdp->data = buf;
		cdp->size = count;
                channel_unlock(chan);
	}

	mem_free(buf, bufsize);
        return 0;
}
#endif /* OSKIT */

static int
chan_edev_init(struct etherdev *edev)
{
	int err;
#ifndef OSKIT
        pthread_t id;
#endif        
        
#ifdef OSKIT        
	/*
	 * Create the netio object we will use for receiving packets.
	 */
	edev->recv = oskit_threaded_netio_create(chan_edev_intr, edev);
	if (edev->recv == 0)
		return ENOMEM; /* XXX */

	/*
	 * Open the device registering our receive netio and getting
	 * back the send netio.
	 */
	assert(edev->send == 0);
	err = oskit_etherdev_open(edev->eth, 0, edev->recv, &edev->send);
	if (err) {
		oskit_netio_release(edev->recv);
		return err;
	}
#else
	edev->recv = os_open_eth(edev->name);
	if (edev->recv < 0) 
		return EINVAL;
        
	err = pthread_create(&id, 0, chan_edev_read_loop, (void *)edev);
	if (err)
		return err;
	err = pthread_detach(id);
	assert(err == 0);
#endif        
	return 0;
}

/*
 * This is where we startup in a flow thread which handles a packet.
 */
static void
chan_eth_receive(void *arg)
{
	channel_t *chan = arg;
	struct chandata *cdata;
	struct chanpdata *cdp, *ecdp;
#ifdef OSKIT
	oskit_bufio_t *b;
	oskit_size_t psize, hsize;
#else
	size_t psize, hsize;
#endif        
	void *data;
	flow_t *flow;
	packet_t *packet;
	channel_deliver_t func;
	int err;
        
	channel_lock(chan);
	cdata = chan->data;
	ecdp = &cdata->pdata[channel_maxthreads(chan)];

	/*
	 * Relocate the bufio.  May not find the same one we were scheduled
	 * for, but it doesn't matter since there should be the same number
	 * of threads running as bufios queued.  Once we find it, we can
	 * drop the channel lock.
	 *
	 * XXX yeah this sucks...
	 */
	for (cdp = cdata->pdata; cdp != ecdp; cdp++)
#ifdef OSKIT
		if (cdp->bufio != 0)
#else
		if (cdp->data != 0)
#endif                        
			break;
	assert(cdp != ecdp);
#ifdef OSKIT
	b = cdp->bufio;
	psize = cdp->size;
        cdp->bufio = 0;
#else
	data = cdp->data;
	psize = cdp->size;
	cdp->data = 0;
#endif        

	flow = chan->flow;
	func = chan->func;
	channel_unlock(chan);

	/*
	 * Copy data into a packet.
	 * XXX should do a bufio/aggregate kinda thing.
	 */
	hsize = ETHER_HDR_SIZE;
	psize -= hsize;
#ifdef OSKIT
	err = oskit_bufio_map(b, &data, hsize, psize);
	if (err) {
		oskit_bufio_release(b);
		return;
	}
	err = channel_packet_create(chan, data, psize, 1, &packet, 0);
	oskit_bufio_unmap(b, data, 0, psize);
	oskit_bufio_release(b);
#else
	/*
	 * Copy the data into a channel receive buffer and free the
	 * passed-in buffer.
	 *
	 * XXX we could use the passed-in buffer directly except that
	 * we have to skip the first 14 bytes which would screw things up
	 * when we want to free/return the buffer.
	 */
	err = channel_packet_create(chan, data + hsize, psize, 1, &packet, 0);
	mem_free(data, psize + hsize);
#endif        
        if (err) 
		return;

	/*
	 * Finally!  We call the application supplied receive function.
	 */
	NODEOS_LEAVE;
	func((an_packet_t)packet);
	NODEOS_ENTER;
}

#endif /* CHAN_ETH */
