/*
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
#include <errno.h>
#include <string.h>

#include "channel.h"
#include "demux.h"
#include "demux_tlist.h"
#include "packet.h"
#include "threadpool.h"
#include "thread.h"
#include "mem.h"
#include "options.h"

#ifdef CHAN_CLOCK
extern void chan_clock_init(void);
extern int chan_clock_create(channel_t *chan);
extern void chan_clock_destroy(channel_t *chan);
#endif
#ifdef CHAN_ETH
extern void chan_eth_init(void);
extern int chan_eth_create(channel_t *chan);
extern void chan_eth_destroy(channel_t *chan);
extern void chan_eth_send(channel_t *chan, packet_t *packet);
#endif
#ifdef CHAN_SOCKET
extern void chan_socket_init(void);
extern int chan_socket_create(channel_t *chan);
extern void chan_socket_destroy(channel_t *chan);
extern void chan_socket_send(channel_t *chan, packet_t *packet);
#endif

static struct chanops allchans[] =
{
#ifdef CHAN_CLOCK
{
	"clock",		chan_clock_init,
	chan_clock_create,	chan_clock_destroy,	0
},
#endif
#ifdef CHAN_ETH
{
	"eth",			chan_eth_init,
	chan_eth_create,	chan_eth_destroy,	chan_eth_send
},
#endif
#ifdef CHAN_SOCKET
{
	"socket",		chan_socket_init,
	chan_socket_create,	chan_socket_destroy,	chan_socket_send
},
#endif
{	0,	0,	0,	0,	0 }
};

void
channel_init(void)
{
	struct chanops *ops;

	for (ops = allchans; ops->name; ops++)
		if (ops->init)
			ops->init();
}

/*
 * Channel this thread is processing for, or zero if none.
 */
channel_t *
channel_current(void)
{
	return thread_current()->chan;
}

int
channel_create(channel_attr_t *attr, demux_tlist_t *demux,
	       channel_deliver_t func, channel_t **out_chan)
{
	channel_t *chan;
	struct chanops *ops;
	int err;

	for (ops = allchans; ops->name; ops++)
		if (strncmp(attr->name, ops->name, strlen(ops->name)) == 0)
			break;

	if (ops->name == 0)
		return EINVAL;

	chan = mem_alloc(sizeof *chan);
	if (chan == 0)
		return ENOMEM;
	memset(chan, 0, sizeof *chan);

	/*
	 * XXX set this now so that functions we call know who to charge
	 * for memory.
	 */
	chan->flow = flow_current();

	chan->tpool = chan->flow->tpool;
	chan->attr = *attr;
	mem_free(attr, sizeof *attr);
	chan->ops = ops;
	chan->demux = demux;
	chan->func = func;
	ilock_init(&chan->lock);

	/*
	 * Allocate receive buffers for in/cut Chans
	 * Note we don't hold the lock here.
	 */
	if (channel_type(chan) != outChan) {
		err = channel_createbufs(chan, channel_maxmemory(chan));
		if (err) {
			ilock_deinit(&chan->lock);
			mem_free(chan, sizeof *chan);
			return err;
		}
	}

	/*
	 * Do channel-type specific setup
	 */
	channel_lock(chan);
	err = ops->create(chan);
	if (err) {
		channel_destroybufs(chan);
		channel_unlock(chan);
		ilock_deinit(&chan->lock);
		mem_free(chan, sizeof *chan);
		return err;
	}

	/*
	 * Finally link to the flow.
	 * We drop the channel lock first to adhere to our locking order.
	 */
	channel_unlock(chan);
	flow_addchannel(chan->flow, chan);

	*out_chan = chan;

	return 0;
}

void
channel_destroy(channel_t *chan)
{
	int me = 0;

	if (channel_current() == chan)
		me = 1;

	/*
	 * XXX fer now band-aid.
	 * Synchronization with channel receive interrupt routines is
	 * a mess right now.  It is bad enough that we don't hold the
	 * channel lock the entire time here, but some interrupt handlers
	 * don't lock it anyway.
	 */
	chan->flags |= CHAN_DYING;

	/*
	 * Remove from the flow before we lock the channel
	 */
	flow_removechannel(chan->flow, chan);

	channel_lock(chan);

#if 0
	/*
	 * Shutdown threads working for the channel
	 */
	if (chan->tpool)
		threadpool_kill(chan->tpool, chan);
#endif

	/*
	 * Channel-specific shutdown
	 */
	chan->ops->destroy(chan);

	/*
	 * Free up all the buffer memory
	 */
	if (chan->rbufs)
		channel_destroybufs(chan);

	if (chan->curthreads == 0) {
		channel_unlock(chan);
		ilock_deinit(&chan->lock);
		mem_free(chan, sizeof *chan);
	} else
		channel_unlock(chan);

	/* XXX this isn't good */
	if (me)
		thread_exit();
}

int
channel_send(channel_t *chan, packet_t *packet)
{
	if (channel_type(chan) != outChan)
		return EINVAL;
	if (chan->flags & CHAN_DYING)
		return EBADF; /* XXX */

	channel_lock(chan);
	chan->curthreads++;
	if (chan->ops->send)
		chan->ops->send(chan, packet);
	chan->curthreads--;
	channel_unlock(chan);

	return 0;
}

int
channel_getaddr(channel_t *chan, channel_addr_t *addr)
{
	if (chan->flags & CHAN_DYING)
		return EBADF; /* XXX */

	channel_lock(chan);
	memcpy(addr, &chan->attr.addr, sizeof chan->attr.addr);
	channel_unlock(chan);
	return 0;
}

void
channel_thread_start(channel_t *chan, packet_t *packet)
{
	int err;

	channel_lock(chan);

	/*
	 * Make sure this channel can pay
	 */
	if (chan->curthreads < channel_maxthreads(chan)) {
		err = thread_start(chan->tpool, chan,
				   (void (*)())chan->func, packet, 0);
		if (err == 0) {
			chan->curthreads++;
			channel_unlock(chan);
			return;
		}
	}
	channel_unlock(chan);
	packet_return(packet);
}

void
channel_thread_exit(channel_t *chan)
{
	channel_lock(chan);
	if (--chan->curthreads == 0 && (chan->flags & CHAN_DYING)) {
		channel_unlock(chan);
		ilock_deinit(&chan->lock);
		mem_free(chan, sizeof *chan);
	} else
		channel_unlock(chan);
}

/*
 * Create one or more buffers an add them to a channel's list of receive
 * buffers.  Called during channel creation.
 */
int
channel_createbufs(channel_t *chan, size_t size)
{
	struct chanbuf *fcb, *cb, *lcb;
	int nbufs;

	/*
	 * Make sure we get at least one buffer out of this
	 */
	nbufs = size / sizeof(struct chanbuf);
	if (nbufs == 0)
		return EINVAL;

	/*
	 * Allocate and chain the new bufs together.
	 * We use the first of the buffer for this.
	 */
	lcb = 0;
	for (fcb = 0; nbufs--; fcb = cb) {
		/* XXX charge the flow/mempool */
		cb = mem_alloc(sizeof *cb);
		if (cb == 0) {
			for (cb = fcb; cb != 0; cb = fcb) {
				fcb = *(struct chanbuf **)cb;
				mem_free(cb, sizeof *cb);
			}
			return ENOMEM;
		}
		*(struct chanbuf **)cb = fcb;
		if (fcb == 0)
			lcb = cb;
	}

	/*
	 * Now add them to the channel's list.
	 * This is done is a separate step to avoid holding the channel
	 * lock too long.
	 */
	channel_lock(chan);
	*(struct chanbuf **)lcb = chan->rbufs;
	chan->rbufs = fcb;
	channel_unlock(chan);

	return 0;
}

void
channel_addbuf(channel_t *chan, struct chanbuf *cbuf)
{
	channel_lock(chan);
	*(struct chanbuf **)cbuf = chan->rbufs;
	chan->rbufs = cbuf;
	channel_unlock(chan);
}

/*
 * Remove and free all buffers associated with a channel.
 * Called during channel destroy.
 * Channel should be locked on call and return.
 */
void
channel_destroybufs(channel_t *chan)
{
	struct chanbuf *fcb, *cb;

	/*
	 * Loop through, freeing all the buffers.
	 * Time spent holding the lock is not an issue here
	 * since the channel is dying anyway.
	 */
	for (cb = chan->rbufs; cb != 0; cb = fcb) {
		fcb = *(struct chanbuf **)cb;
		mem_free(cb, sizeof *cb);
	}
}

/*
 * XXX note: assumes all buffers are the same size
 */
int
channel_packet_create(channel_t *chan, void *data, int dlen, int mustcopy,
		      packet_t **out_packet, struct chanbuf **out_buf)
{
	struct chanbuf *newbuf;
	packet_t *packet;
	int err;

	channel_lock(chan);

	/*
	 * Channel is shutting down
	 */
	if (chan->flags & CHAN_DYING) {
		channel_unlock(chan);
		return EINVAL;
	}

	/*
	 * Ensure the channel has resources for this packet
	 */
	newbuf = chan->rbufs;
	if (newbuf == 0) {
		channel_unlock(chan);
		return ENOMEM;
	}
	chan->rbufs = *(struct chanbuf **)newbuf;
	
	/*
	 * Copy the packet if necessary
	 */
	if (mustcopy) {
		void *tmp;

		memcpy(newbuf, data, dlen);
		tmp = data;
		data = newbuf;
		newbuf = tmp;
	}

	/*
	 * Create a packet object from the buffer.
	 */
	err = packet_create_external(chan->flow, chan, data, dlen, mem_free,
				     sizeof(struct chanbuf), &packet);
	if (err) {
		if (mustcopy)
			newbuf = data;
		*(struct chanbuf **)newbuf = chan->rbufs;
		chan->rbufs = newbuf;
		channel_unlock(chan);
		return err;
	}
	channel_unlock(chan);

	*out_packet = packet;
	if (out_buf)
		*out_buf = newbuf;
	return 0;
}

/*
 * Create and install a demux object on the given demux list.
 * Channel should be locked on call.
 */
int
channel_demux_add(channel_t *chan, struct demux **listp)
{
	demux_t *demux;
	int prio, err;

	if (chan->demux) {
		err = demux_create(chan, demux_tlist_match, chan->demux,
				   &demux);
		prio = chan->demux->ntriples; /* XXX */
	} else {
		err = demux_create(chan, 0, 0, &demux);
		prio = 0; /* XXX */
	}

	if (err == 0) {
		err = demux_add(listp, demux, prio);
		if (err)
			demux_destroy(demux);
	}

	return err;
}

void
channel_demux_remove(channel_t *chan, struct demux **listp)
{
	demux_t *demux;

	demux = demux_find(*listp, chan);
	if (demux) {
		demux_remove(listp, demux);
		demux_destroy(demux);
	}
}

