/*
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
#include <errno.h>
#include <string.h>
#include <assert.h>

#include "demux_tlist.h"
#include "demux.h"
#include "mem.h"

static struct demux_triple *demux_triple_create(int off, int len, void *key);
static void demux_triple_destroy(struct demux_triple *dt);

/*
 * Create a triple list large enough for the indicated number of elements
 */
int
demux_tlist_create(int ntriples, demux_tlist_t **out_list)
{
	demux_tlist_t *list;
	size_t size;

	size = sizeof *list + (ntriples * sizeof(struct demux_triple *));
	list = mem_alloc(size);
	if (list == 0)
		return ENOMEM;

	memset(list, 0, size);
	list->ntriples = ntriples;

	*out_list = list;
	return 0;
}

/*
 * Create and add a match element to the front of the given list.
 */
int
demux_tlist_prepend(demux_tlist_t *list, int off, int len, void *key)
{
	struct demux_triple *dt;
	int i;

	for (i = 0; i < list->ntriples; i++)
		if (list->list[i] == 0)
			break;

	if (i == list->ntriples)
		return ENOMEM;

	dt = demux_triple_create(off, len, key);
	if (dt == 0)
		return ENOMEM;

	while (i > 0) {
		list->list[i] = list->list[i-1];
		i--;
	}
	list->list[i] = dt;

	return 0;
}


/*
 * Create and add a match element to the end of the given list.
 */
int
demux_tlist_append(demux_tlist_t *list, int off, int len, void *key)
{
	struct demux_triple *dt;
	int i;

	for (i = 0; i < list->ntriples; i++)
		if (list->list[i] == 0)
			break;

	if (i == list->ntriples)
		return ENOMEM;

	dt = demux_triple_create(off, len, key);
	if (dt == 0)
		return ENOMEM;

	list->list[i] = dt;

	return 0;
}

/*
 * Copy the contents of an existing list into a possibly larger new list.
 */
int
demux_tlist_copy(demux_tlist_t *list, int ntriples, demux_tlist_t **out_list)
{
	demux_tlist_t *newlist;
	int err;

	assert(ntriples >= list->ntriples);
	err = demux_tlist_create(ntriples, &newlist);
	if (err == 0) {
		memcpy(newlist->list, list->list,
		       list->ntriples * sizeof(struct demux_triple *));
		*out_list = newlist;
	}

	return err;
}

/*
 * Adjust the offset of all triples by the indicated amount
 */
void
demux_tlist_adjust(demux_tlist_t *list, int adjust)
{
	int i;

	for (i = 0; i < list->ntriples; i++)
		if (list->list[i])
			list->list[i]->off += adjust;
}

/*
 * Destroy a triple list
 */
void
demux_tlist_destroy(demux_tlist_t *list)
{
	int i;
	size_t size;

	size = sizeof *list + (list->ntriples * sizeof(struct demux_triple *));
	for (i = 0; i < list->ntriples; i++)
		if (list->list[i])
			demux_triple_destroy(list->list[i]);

	mem_free(list, size);
}

/*
 * Demux match function for a triple list
 */
int
demux_tlist_match(void *ddata, void *pdata, int plen)
{
	demux_tlist_t *list = ddata;
	struct demux_triple *dt;
	int i;

	for (i = 0; i < list->ntriples; i++) {
		dt = list->list[i];
		if (dt == 0)
			break;

#if 0
printf("m%p: %d/%d/%x... to %d/%x...\n",
dt, dt->off, dt->len, *(unsigned *)dt->key, plen, *(unsigned *)(pdata+dt->off));
#endif
		if (dt->off + dt->len >= plen ||
		    memcmp(pdata+dt->off, dt->key, dt->len) != 0)
			return 0;
	}

	return 1;
}

static struct demux_triple *
demux_triple_create(int off, int len, void *key)
{
	struct demux_triple *dt;

	dt = mem_alloc(sizeof *dt + len);
	if (dt) {
		dt->off = off;
		dt->len = len;
		if (key)
			memcpy(dt->key, key, len);
		else
			memset(dt->key, 0, len);
	}

	return dt;
}

static void
demux_triple_destroy(struct demux_triple *dt)
{
	mem_free(dt, sizeof *dt + dt->len);
}
