/*
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * EE to test flow termination.
 */
#include "options.h"

#ifdef EE_FLOWTEST

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <pthread.h>

#include <nodeos/impl.h>
#include <nodeos/flow.h>
#include <nodeos/thread.h>
#include <nodeos/synch.h>

#define NUMFLOWS	10
#define NUMTHREADS	3

static void ee_init(void *arg);
an_flowinit_t InitEE = { ee_init };

static void ee_spin(void *arg);
static void ee_syscall(void *arg);
static void ee_block(void *arg);
static void ee_child(void *arg);
static void ee_death(void *arg);

int level = 0;
static an_flow_t ee_startchild(void);
static an_mutex_t lock;
static an_cond_t forever;

static an_mutex_t threadlock;
static an_cond_t threadcond;
static an_thread_t mythreads[NUMFLOWS*NUMTHREADS];
static int nthread;

int
ee_main(int argc, char **argv)
{
	an_flow_t iflow;
	an_timespec_t stime;

	printf("%p/%p: Flow test EE started\n",
	       an_flow_current(), an_thread_current());

	lock = an_mutex_create();
	forever = an_cond_create();

	threadlock = an_mutex_create();
	threadcond = an_cond_create();

	printf("%p/%p: Creating child flow...\n",
	       an_flow_current(), an_thread_current());
	iflow = ee_startchild();

	an_mutex_lock(threadlock);
	while (nthread != NUMTHREADS * NUMFLOWS)
		an_cond_wait(threadcond, threadlock);
	an_mutex_unlock(threadlock);

	stime.tv_sec = 3;
	stime.tv_nsec = 0;
	an_thread_sleep(&stime);

	printf("%p/%p: Killing child flow %p!\n",
	       an_flow_current(), an_thread_current(), iflow);
	an_flow_destroy(iflow);

	return 0;
}

static an_flow_t
ee_startchild(void)
{
	an_flow_t flow;
	an_cpuspec_t cs;
	an_memspec_t ms;
	an_flowinit_t fss;

	/* XXX all that matters right now */
	memset(&cs, 0, sizeof cs);
	cs.max = NUMTHREADS+1;
	cs.stacksize = 20 * 1024;
	ms.npages = 100;
	ms.pool = an_mempool_current();
	fss = ee_child;
	flow = an_flow_create(an_cred_current(), &cs, &ms, fss, ee_death);
	assert(flow != 0);

	return (flow);
}

static void
ee_child(void *arg)
{
	an_error_t err;
	int tix;

	tix = level * NUMTHREADS;

	an_mutex_lock(threadlock);
	printf("%p/%p: Firing off my infinite loop thread...\n",
	       an_flow_current(), an_thread_current());
	err = an_thread_start(&mythreads[tix+0], an_flow_current(),
			      ee_spin, 0);
	assert(err == 0);

	printf("%p/%p: Firing off my syscall thread...\n",
	       an_flow_current(), an_thread_current());
	err = an_thread_start(&mythreads[tix+1], an_flow_current(),
			      ee_syscall, 0);
	assert(err == 0);

	printf("%p/%p: Firing off my mutex-wait thread...\n",
	       an_flow_current(), an_thread_current());
	err = an_thread_start(&mythreads[tix+2], an_flow_current(),
			      ee_block, 0);
	assert(err == 0);

	printf("%p/%p: threads %p/%p/%p started\n",
	       an_flow_current(), an_thread_current(),
	       mythreads[tix+0], mythreads[tix+1], mythreads[tix+2]);
	fflush(stdout);
	an_mutex_unlock(threadlock);

	if (++level < NUMFLOWS) {
		printf("%p/%p: Firing off a sub flow...\n",
		       an_flow_current(), an_thread_current());
		(void) ee_startchild();
	}
}

static void
ee_ihandler(int itype, void *info)
{
	an_mutex_lock(threadlock);
	printf("%p/%p: %s, exiting\n", an_flow_current(), an_thread_current(),
	       itype == AN_THREAD_TRAP ? "TRAP" : "INTR");
	nthread--;
	an_cond_signal(threadcond);
	an_mutex_unlock(threadlock);
	an_thread_exit();
}

static void
ee_ihandler2(int itype, void *info)
{
	an_mutex_lock(threadlock);
	if (an_mutex_trylock(lock) == 0)
		printf("%p/%p: cond_wait inthandler: mutex not locked!\n",
		       an_flow_current(), an_thread_current());
	printf("%p/%p: %s(cond_wait), exiting\n",
	       an_flow_current(), an_thread_current(),
	       itype == AN_THREAD_TRAP ? "TRAP" : "INTR");
	nthread--;
	an_cond_signal(threadcond);
	an_mutex_unlock(threadlock);
	an_mutex_unlock(lock);
	an_thread_exit();
}

static void
ee_spin(void *arg)
{
	an_mutex_lock(threadlock);
	nthread++;
	printf("%p/%p: SPIN thread (%d)\n",
	       an_flow_current(), an_thread_current(), nthread);
	an_thread_sethandler(ee_ihandler);
	an_cond_signal(threadcond);
	an_mutex_unlock(threadlock);

	while (1)
		;
}

static void
ee_syscall(void *arg)
{
	an_mutex_lock(threadlock);
	nthread++;
	printf("%p/%p: SYSCALL thread (%d)\n",
	       an_flow_current(), an_thread_current(), nthread);
	an_thread_sethandler(ee_ihandler);
	an_cond_signal(threadcond);
	an_mutex_unlock(threadlock);

	while (1)
		(void) an_flow_current();
}

static void
ee_block(void *arg)
{
	an_mutex_lock(threadlock);
	nthread++;
	printf("%p/%p: BLOCKED thread (%d)\n",
	       an_flow_current(), an_thread_current(), nthread);
	an_thread_sethandler(ee_ihandler2);
	an_cond_signal(threadcond);
	an_mutex_unlock(threadlock);

	an_mutex_lock(lock);
	while (1)
		an_cond_wait(forever, lock);
}

static void
ee_death(void *arg)
{
	an_flow_t flow = arg;
	int i;

	printf("%p/%p: NullEE: flow %p killed, interrupting %d threads\n",
	       an_flow_current(), an_thread_current(), flow, nthread);
	an_mutex_lock(threadlock);
	for (i = 0; i < nthread; i++)
		an_thread_intr(mythreads[i]);
	while (nthread > 0)
		an_cond_wait(threadcond, threadlock);
	printf("%p/%p: NullEE: all threads gone!\n",
	       an_flow_current(), an_thread_current());
	an_mutex_unlock(threadlock);
}

static void
ee_init(void *arg)
{
	int status;
	int argc = 1;
	char *argv[] = { "FlowtestEE" }; 
	an_flow_t fid = an_flow_current();

	status = ee_main(argc, argv);
	an_flow_destroy(fid);
}

#endif
