/*
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include "resource.h"
#include "credential.h"
#include "threadpool.h"
#include "thread.h"
#include "mempool.h"
#include "mem.h"
#include "flow.h"
#include "channel.h"

flow_t *root_flow;
static int flow_cleanup(struct hnode *h, void *arg);
static void flow_upcall(flow_t *fflow, flow_t *tflow);

void
flow_init(flow_init_t initfunc, flow_term_t termfunc)
{
	extern struct credential *root_cred;
	extern struct mempool *root_mempool;
	flow_t *flow;
	cpuspec_t cspec;
	memspec_t mspec;
	int err;

	cspec.max = cspec.rate = RESOURCE_ALL;
	cspec.slice.tv_sec = RESOURCE_ALL;
	cspec.sched = 0; /* XXX */
	cspec.stacksize = 64 * 1024;
	mspec.npages = RESOURCE_ALL;
	mspec.pool = (an_mempool_t)root_mempool;
	err = flow_create(root_cred, &cspec, &mspec, initfunc, termfunc, &flow);
	assert(err == 0);
}

flow_t *
flow_current(void)
{
	thread_t *thread = thread_current();

	return thread ? thread->pool->flow : 0;
}

int flow_create(struct credential *cred, cpuspec_t *cspec, memspec_t *mspec,
		flow_init_t initfunc, flow_term_t termfunc,
		struct flow **out_flow)
{
	flow_t *flow, *pflow;
	threadpool_t *tpool;
	resource_t res;
	int err;

	/* Sanity */
	if (cred == 0 || cspec == 0 || mspec == 0 ||
	    initfunc == 0 || termfunc == 0 || out_flow == 0)
		return EINVAL;

	/* Can this flow create sub flows? */
	if (credential_check(cred, CCHECK_FLOW_CREATE) == CCHECK_FAIL)
		return EPERM;

	/* Ensure requested resources are available and reserve them */
	resource_init(&res, cspec, mspec);
	err = credential_resource_alloc(cred, &res);
	if (err)
		return err;

	pflow = flow_current();

	/* Allocate and initialize the flow and threadpool structs */
	flow = mem_alloc(sizeof *flow);
	if (flow == 0) {
		credential_resource_free(cred, &res);
		return ENOMEM;
	}
	err = threadpool_create(flow, 0, cspec, &tpool);
	if (err) {
		mem_free(flow, sizeof *flow);
		credential_resource_free(cred, &res);
		return err;
	}
	hnode_init(&flow->me);
	flow->cred = cred;
	flow->res = res;
	flow->tpool = tpool;
	flow->chanlist = 0;
	flow->flags = 0;
	flow->termfunc = termfunc;
	mutex_init(&flow->lock);

	/* Link into flow hierarchies (special handling for first flow) */
	if (root_flow == 0) {
		root_flow = flow;
		flow_lock(flow);
		hnode_attach(&flow->me, 0, 0);
	} else {
		flow_lock(root_flow);
		flow_lock(flow);
		hnode_attach(&flow->me, &flow_current()->me, 0);
		flow_unlock(root_flow);
	}

	/* Link onto the mempool list */
	flow->mpool = (mempool_t *)mspec->pool;
	mempool_addflow(flow->mpool, flow);

	flow_unlock(flow);

	/* All setup, fire up a thread in the flow to call the init function */
	if (initfunc) {
		err = thread_start(tpool, 0, initfunc, 0, 0);
		if (err) {
			flow_destroy(flow);
			return err;
		}
	}

	*out_flow = flow;
	return 0;
}

int
flow_destroy(flow_t *flow)
{
	flow_t *oflow;
	int me = 0, err;

	assert(flow != root_flow);

	/*
	 * Flow must be us or one of our immediate children
	 */
	if (flow == flow_current())
		me = 1;
	else {
		for (oflow = flow_child(flow_current()); oflow != 0;
		     oflow = flow_sibling(oflow))
			if (oflow == flow)
				break;
		if (oflow == 0)
			return EINVAL;
	}

	/*
	 * Prevent recursive upcalls and, effectively, attempts to destroy
	 * an already dying flow.  Note that the flow lock is not held while
	 * in the upcall so we need this other mechanism to ensure this.
	 *
	 * XXX we need to do this better however since there is a race from
	 * here to when the upcall is actually made.
	 */
	if (flow->flags & FLOW_INUPCALL)
		return 0;

	oflow = flow_parent(flow);
	flow_upcall(flow, oflow);

	if (flow != root_flow)
		flow_lock(root_flow);

	err = hnode_traverse(&flow->me, flow_cleanup, 0, HNODE_PARENT, 1);
	assert(err == 0);

	if (flow != root_flow)
		flow_unlock(root_flow);

	/* If we just destroyed ourself, no where to return to */
	if (me) {
		/* XXX what resources are we running on? */
		thread_exit();
	}

	return 0;
}

/*
 * The guts of flow destruction.  Called by the hnode_traverse in
 * flow_destroy.
 */
static int
flow_cleanup(struct hnode *h, void *arg)
{
	flow_t *flow = (flow_t *)h;
	flow_t *pflow;
	channel_t *chan, *next;

	flow_lock(flow);
	pflow = flow_parent(flow);
	flow->flags |= FLOW_DYING;

	/* Remove from flow hierarchy */
	hnode_detach(h, HNODE_PARENT);

	/* Destroy all the channels */
	for (chan = flow->chanlist; chan != 0; chan = next) {
		next = chan->next;
		channel_destroy(chan);
	}

	/* Kill any remaining threads */
	if (flow->tpool)
		threadpool_destroy(flow->tpool);

	/* Remove from mempool */
	assert(hnode_listhead(&flow->me) != 0);
	mempool_removeflow(flow->mpool, flow);

	credential_resource_free(flow->cred, &flow->res);

	flow_unlock(flow);

	mutex_deinit(&flow->lock);
	mem_free(flow, sizeof *flow);

	return 0;
}

/*
 * Upcall to a flow's parent termination handler.
 * If there are no resources to make the call (or the call takes too long)
 * that is fine, we will just forcibly kill everything. 
 */
static void
flow_upcall(flow_t *flow, flow_t *parent)
{
	if (flow->termfunc == 0)
		return;

	flow->flags |= FLOW_INUPCALL; /* XXX races */

	assert(parent->tpool != 0);
	(void) thread_run(parent->tpool, flow->termfunc, flow, 0);
}

void
flow_addchannel(flow_t *flow, channel_t *chan)
{
	flow_lock(flow);
	channel_lock(chan);

	/* XXX make sure requested channel resources are consistent with flow */

	chan->flow = flow;
	chan->next = flow->chanlist;
	flow->chanlist = chan;

	channel_unlock(chan);
	flow_unlock(flow);
}

void
flow_removechannel(flow_t *flow, channel_t *chan)
{
	channel_t *ch, **pch;

	assert(chan->flow == flow);

	/*
	 * Flows never return from the dead so we don't lock this access
	 */
	if (flow->flags & FLOW_DYING)
		return;

	flow_lock(flow);
	channel_lock(chan);

	for (pch = &flow->chanlist; (ch = *pch) != 0; pch = &ch->next)
		if (ch == chan) {
			*pch = ch->next;
			break;
		}
	chan->flow = 0;

	channel_unlock(chan);
	flow_unlock(flow);
}

#include <stdio.h>

static int
dump_flow(struct hnode *node, void *arg)
{
	flow_t *flow = (flow_t *)node;

	printf("FLOW %p: parent=%p, child=%p, sibling=%p, "
	       "cred=%p, mpool=%p, tpool=%p\n",
	       flow, hnode_parent(node), hnode_child(node), hnode_sibling(node),
	       flow->cred, flow->mpool, flow->tpool);
	return 0;
}

void
flow_dump(flow_t *flow)
{
	if (flow == 0)
		flow = flow_current();

	hnode_traverse(&flow->me, dump_flow, 0, HNODE_PARENT, 0);
}
