/*
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <assert.h>
#include <stdlib.h>
#include <errno.h>

#include "hnode.h"
#include "mem.h"
#include "mempool.h"
#include "flow.h"

/*
 */

static struct hnode allpools;
struct mempool *root_mempool;

void
mempool_init(void)
{
	int err;

	err = mempool_create(0, 0, &root_mempool);
	assert(err == 0);
}

/*
 * Get the mempool associated with the current thread's flow.
 */
mempool_t *
mempool_current(void)
{
	flow_t *flow = flow_current();

	return flow ? flow->mpool : 0;
}

/*
 * Create a new mempool with no associated flows.  The parent
 * parameter defines this pool's position in the pool hierarchy.
 * The callback routine is invoked whenever the NodeOS detects that
 * the indicated pool is over its page limit.
 */
int
mempool_create(mempool_t *parent, mempool_cbfunc_t callback,
	       mempool_t **out_pool)
{
	mempool_t *pool;

	/* XXX pool storage charged to caller flow */
	pool = mem_alloc(sizeof *pool);
	if (pool == 0)
		return ENOMEM;

	hnode_init(&pool->me);
	hnode_attach(&pool->me, 0, &allpools);
	if (parent)
		hnode_attach(&pool->me, (struct hnode *)parent, 0);
	pool->physcurrent = pool->physlimit = 0;
	pool->callback = callback;

	*out_pool = pool;
	return 0;
}

/*
 * Destroy a mempool.  There should be no flows associatd with the pool.
 */
int
mempool_destroy(mempool_t *pool)
{
	if (hnode_child(&pool->me) != 0)
		return EBUSY;

	hnode_detach(&pool->me, HNODE_ALL);

	mem_free(pool, sizeof *pool);

	return 0;
}

/*
 * Donates the memory resources associated with a flow to the pool.
 */
void
mempool_addflow(mempool_t *pool, flow_t *flow)
{
	hnode_attach((struct hnode *)flow, 0, &pool->flowlist);

	pool->physlimit += flow->res.ppages;

#if 0
	/*
	 * XXX first attach to the root pool,
	 * debit for memory allocated before this point.
	 */
	if (pool == root_mempool && pool->physcurrent == 0) {
		extern flow_t *root_flow;
		extern unsigned long earlymempages;

		assert(flow == root_flow);
		assert(earlymempages <= pool->physlimit);

		pool->physcurrent += earlymempages;
	}
#endif
}

/*
 * Removes the memory resources associated with a flow from the pool.
 */
void
mempool_removeflow(mempool_t *pool, flow_t *flow)
{
	hnode_detach((struct hnode *)flow, HNODE_GLOBAL);

	pool->physlimit -= flow->res.ppages;

#if 0
	/*
	 * If we are over the pool limit and it matters, invoke the callback
	 * to free up some memory.  If no callback, kill all the flows!
	 *
	 * XXX we need to establish the proper context for the callback to
	 *     run in (whatever that might be)
	 * XXX we need to do something reasonable if the callback doesn't
	 *     return or doesn't free memory.
	 */
	if (callback == 0)
		;
	while (pool->physcurrent > pool->physlimit)
		pool->callback(pool, pool->physcurrent, pool->physlimit);
#endif
}

/*
 * Simple malloc/free allocation interface
 *
 * XXX if pool == 0, root_mempool does not yet exist or has no resources yet,
 * remember allocations for later.
 *
 * XXX maintain seperate heaps for each pool, charge accordingly
 */

void *
mempool_alloc(mempool_t *pool, size_t size)
{
	/* XXX fer now */
	return malloc(size);
}

void
mempool_free(mempool_t *pool, void *ptr, size_t size)
{
	/* XXX fer now */
	free(ptr);
}


#include <stdio.h>

static int
dump_mempool(struct hnode *node, void *arg)
{
	mempool_t *pool = (mempool_t *)node;

	printf("POOL %p: parent=%p, child=%p, sibling=%p, physcur=%d, physlimit=%d\n",
	       pool, hnode_parent(node), hnode_child(node),
	       hnode_sibling(node), pool->physcurrent, pool->physlimit);
	return 0;
}

void
mempool_dump(mempool_t *pool)
{
	if (pool == 0)
		hnode_traverse(&mempool_current()->me, dump_mempool, 0,
			       HNODE_GLOBAL, 0);
	else
		hnode_traverse(&pool->me, dump_mempool, 0, HNODE_PARENT, 0);
}
