/*
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <errno.h>
#include <assert.h>

#include "synch.h"
#include "thread.h"
#include "mem.h"

/*
 * N.B. NodeOS mutexes are not cancelable.
 * This would obviously have to change if the NodeOS doesn't trust the EE.
 */

void
mutex_init(mutex_t *mutex)
{
	int err;

	err = pthread_mutex_init(&mutex->mutex, 0);
	assert(err == 0);
}

void
mutex_deinit(mutex_t *mutex)
{
	int err;

	err = pthread_mutex_destroy(&mutex->mutex);
	assert(err == 0);
}

int
mutex_create(mutex_t **out_mutex)
{
	mutex_t *mutex;
	int err;

	mutex = mem_alloc(sizeof *mutex);
	if (mutex == 0)
		return ENOMEM;

	err = pthread_mutex_init(&mutex->mutex, 0);
	if (err) {
		mem_free(mutex, sizeof *mutex);
		return EINVAL;
	}

	*out_mutex = mutex;
	return 0;
}

void
mutex_lock(mutex_t *mutex)
{
	pthread_mutex_lock(&mutex->mutex);
}

int
mutex_trylock(mutex_t *mutex)
{
	return pthread_mutex_trylock(&mutex->mutex);
}

void
mutex_unlock(mutex_t *mutex)
{
	pthread_mutex_unlock(&mutex->mutex);
}

int
mutex_destroy(mutex_t *mutex)
{
	int err;

	err = pthread_mutex_destroy(&mutex->mutex);
	if (err)
		return err;

	mem_free(mutex, sizeof *mutex);
	return 0;
}


void
cond_init(cond_t *cond)
{
	int err;

	ilock_init(&cond->lock);
	err = pthread_cond_init(&cond->cond, 0);
	assert(err == 0);
}

void
cond_deinit(cond_t *cond)
{
	int err;

	err = pthread_cond_destroy(&cond->cond);
	assert(err == 0);
	ilock_deinit(&cond->lock);
}

int
cond_create(cond_t **out_cond)
{
	cond_t *cond;
	int err;

	cond = mem_alloc(sizeof *cond);
	if (cond == 0)
		return ENOMEM;

	err = pthread_cond_init(&cond->cond, 0);
	if (err) {
		mem_free(cond, sizeof *cond);
		return EINVAL;
	}
	ilock_init(&cond->lock);

	*out_cond = cond;
	return 0;
}

void
cond_wait(cond_t *cond, mutex_t *mutex)
{
	cond_lock(cond);

	/*
	 * Release the mutex and wait.
	 * Note that we are still atomic here since no one can signal the
	 * condition until we release the cond lock.
	 */
	mutex_unlock(mutex);
	thread_condwait(cond);

	/*
	 * Now unlock the condition before attempting to
	 * reacquire the mutex.  There is no race here.
	 */
	cond_unlock(cond);
	mutex_lock(mutex);
}

void
cond_signal(cond_t *cond)
{
	int err;

	cond_lock(cond);
	err = pthread_cond_signal(&cond->cond);
	assert(err == 0);
	cond_unlock(cond);
}

void
cond_broadcast(cond_t *cond)
{
	int err;

	cond_lock(cond);
	err = pthread_cond_broadcast(&cond->cond);
	assert(err == 0);
	cond_unlock(cond);
}

int
cond_destroy(cond_t *cond)
{
	cond_deinit(cond);
	mem_free(cond, sizeof *cond);

	return 0;
}
