/*
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <sched.h>

#include "nodeos.h"
#include "flow.h"
#include "thread.h"
#include "threadpool.h"
#include "channel.h"
#include "mem.h"
#include "misc.h"

static pthread_key_t nodeos_thread;	/* thread struct for this pthread */
static pthread_key_t nodeos_threaddata;	/* nodeos exported thread data */

extern void os_dump_state(void *state);
extern void panic(const char *fmt, ...);


/*
 * For hack thread_sleep implementation.
 * Cond is never signalled, just allows us to do a timed wait.
 */
static pthread_mutex_t sleepmutex;
static pthread_cond_t sleepcond;

static void *thread_startup(void *arg);
static int thread_start_common(threadpool_t *tpool, channel_t *chan, int flags,
			       void (*func)(void *arg), void *arg,
			       thread_t **out_thread);
static void thread_cleanup(void *arg);
static void thread_intr_init(void);

/*
 * Auxilliary struct for passing data to a new pthread in thread_startup.
 */
struct threaddata {
	thread_t *thread;	/* The NodeOS identity */
	threadpool_t *tpool;	/* its threadpool */
	int flags;		/* flags */
	void (*func)(void *);	/* function the thread should start in */
	void *arg;		/* and and argument for that function */
};

void
thread_init(void)
{
	int err;

	/*
	 * Establish a thread_specific data key to hold the thread_t
	 * associated with a pthread.  This is used to determine the
	 * current thread.  Init to zero for this primortial thread.
	 */
	err = pthread_key_create(&nodeos_thread, 0);
	assert(err == 0);
	err = pthread_setspecific(nodeos_thread, 0);
	assert(err == 0);

	/*
	 * Also need a key to use for the exported thread_setdata
	 * interface.
	 */
	err = pthread_key_create(&nodeos_threaddata, 0);
	assert(err == 0);
	err = pthread_setspecific(nodeos_threaddata, 0);
	assert(err == 0);

	/*
	 * For thread sleep
	 */
	err = pthread_mutex_init(&sleepmutex, 0);
	assert(err == 0);
	err = pthread_cond_init(&sleepcond, 0);
	assert(err == 0);

	thread_intr_init();
}

thread_t *
thread_current(void)
{
	/*
	 * In the current implementation has no control over scheduling
	 * of threads, so we have to query the pthread state.
	 */
	return pthread_getspecific(nodeos_thread);
}

/*
 * Called during thread pool initialization to ensure that thread parameters
 * are compatible with the underlying thread implementation.
 */
int
thread_checkattrs(cpuspec_t *cspec)
{
	pthread_attr_t attr;
	int err;

	err = pthread_attr_init(&attr);
	assert(err == 0);
	err = pthread_attr_setstacksize(&attr, cspec->stacksize);
	return err;
}

/*
 * Start up a thread from a specific thread pool.
 * Returns thread identifier if desired.
 */
int
thread_start(threadpool_t *tpool, channel_t *chan,
	     void (*func)(void *arg), void *arg, thread_t **out_thread)
{
	thread_t *thread;
	int err;

	threadpool_lock(tpool);
	err = thread_start_common(tpool, chan, 0, func, arg, &thread);
	if (err) {
		threadpool_unlock(tpool);
		return err;
	}

	err = pthread_detach(thread->id);
	assert(err == 0);

	threadpool_unlock(tpool);

	if (out_thread)
		*out_thread = thread;
	return 0;
}

/*
 * Like thread_start but start thread in NodeOS instead of a ``user'' function.
 */
int
thread_start_kernel(threadpool_t *tpool, channel_t *chan,
		    void (*func)(void *arg), void *arg, thread_t **out_thread)
{
	thread_t *thread;
	int err;

	threadpool_lock(tpool);
	err = thread_start_common(tpool, chan, THREAD_INKERNEL, func, arg,
				  &thread);
	if (err) {
		threadpool_unlock(tpool);
		return err;
	}

	err = pthread_detach(thread->id);
	assert(err == 0);

	threadpool_unlock(tpool);

	if (out_thread)
		*out_thread = thread;
	return 0;
}

/*
 * Like thread_start but await completion
 *
 * XXX implement timeouts
 */
int
thread_run(threadpool_t *tpool, void (*func)(void *arg), void *arg,
	   an_timespec_t *timo)
{
	thread_t *thread;
	int err;

	threadpool_lock(tpool);
	err = thread_start_common(tpool, 0, 0, func, arg, &thread);
	if (err) {
		threadpool_unlock(tpool);
		return err;
	}

	thread->flags |= THREAD_WAITING;
	while ((thread->flags & THREAD_WAITING) != 0)
		icond_wait(&thread->waiting, &tpool->lock);

	ilock_deinit(&thread->lock);
	icond_deinit(&thread->waiting);
	mem_free(thread, sizeof *thread);

	/*
	 * Cleanup the thread_pool if necessary
	 */
	if (tpool->curthreads == 0)
		threadpool_cleanup(tpool);
	else
		threadpool_unlock(tpool);

	return 0;
}

static int
thread_start_common(threadpool_t *tpool, channel_t *chan, int flags,
		    void (*func)(void *arg), void *arg, thread_t **out_thread)
{
	thread_t *thread;
	pthread_attr_t attr;
	struct threaddata *tdata;
	int err;

	if (tpool->curthreads >= tpool->params.max)
		return EAGAIN;

	thread = mem_alloc(sizeof *thread);
	if (thread == 0)
		return ENOMEM;
	memset(thread, 0, sizeof *thread);

	/* Set up arguments for the new thread */
	tdata = mem_alloc(sizeof *tdata);
	if (tdata == 0) {
		mem_free(thread, sizeof *thread);
		return ENOMEM;
	}
	tdata->thread = thread;
	tdata->tpool = tpool;
	tdata->flags = flags;
	tdata->func = func;
	tdata->arg = arg;

	/*
	 * Set up the attributes for the thread.
	 * Currently we only set the stack size.
	 * (XXX and global scope for the benefit of Solaris)
	 */
	err = pthread_attr_init(&attr);
	assert(err == 0);
	err = pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
	assert(err == 0);
	err = pthread_attr_setstacksize(&attr, tpool->params.stacksize);
	assert(err == 0);

	/*
	 * Create the pthread, it will block immediately on the pool lock.
	 */
	err = pthread_create(&thread->id, &attr, thread_startup, tdata);
	if (err) {
		mem_free(tdata, sizeof *tdata);
		mem_free(thread, sizeof *thread);
		return err;
	}

	/* We are committed, finish up */
	ilock_init(&thread->lock);
	thread->flags = THREAD_INKERNEL;
	thread->chan = chan;
	thread->pool = tpool;
	thread->next = tpool->threads;
	icond_init(&thread->waiting);
	tpool->threads = thread;
	tpool->curthreads++;

	if (out_thread)
		*out_thread = thread;

	return 0;
}

static void *
thread_startup(void *arg)
{
	struct threaddata *tdata = arg;
	thread_t *thread;
	threadpool_t *tpool;
	void (*func)(void *);
	void *farg;
	int flags;

	/* Record our identity */
	thread = tdata->thread;
	pthread_setspecific(nodeos_thread, thread);

	/* Grab the other arguments */
	tpool = tdata->tpool;
	flags = tdata->flags;
	func = tdata->func;
	farg = tdata->arg;

	/* Grab the thread pool lock so we know when our creator is done */
	threadpool_lock(tpool);
	threadpool_unlock(tpool);

	/*
	 * Free the memory now that we officially belong to a pool
	 * (determining our current flow requires indirect through pool).
	 */
	mem_free(tdata, sizeof *tdata);

	/*
	 * Now that we are official, push a cleanup handler, mark the thread
	 * as interruptible if it is returning to "user mode", and launch it.
	 */
	pthread_cleanup_push(thread_cleanup, thread);
	if ((flags & THREAD_INKERNEL) == 0) {
		thread->flags |= THREAD_INTSON;
		NODEOS_LEAVE;
		func(farg);
		NODEOS_ENTER;
	} else
		func(farg);

	/*
	 * Call pthread_exit explicitly here instead of returning
	 * since the cleanup handler must be "in scope" when we exit.
	 */
	pthread_exit(0);
	pthread_cleanup_pop(1);

	return 0;
}

void
thread_kill(thread_t *thread)
{
	int err;

	assert(thread != thread_current());

	err = pthread_cancel(thread->id);
	assert(err == 0);
}

void
thread_exit(void)
{
	/* Just exit, the actual work is done in the cleanup handler */
	pthread_exit(0);
}

static void
thread_cleanup(void *arg)
{
	thread_t *thread = arg;
	thread_t **pth, *th;
	threadpool_t *tpool;

	assert(thread == thread_current());

	if (thread->chan)
		channel_thread_exit(thread->chan);

	tpool = thread->pool;
	threadpool_lock(tpool);
	for (pth = &tpool->threads; (th = *pth) != 0; pth = &th->next)
		if (th == thread)
			break;
	assert(th);
	*pth = thread->next;

	tpool->curthreads--;
	/*
	 * Can't cleanup the threadpool in the waiting case since
	 * the waiter will need to reacquire its lock in thread_run.
	 */
	if (thread->flags & THREAD_WAITING) {
		thread->flags &= ~THREAD_WAITING;
		icond_signal(&thread->waiting);
		threadpool_unlock(tpool);
	} else {
		ilock_deinit(&thread->lock);
		icond_deinit(&thread->waiting);
		mem_free(thread, sizeof *thread);
		if (tpool->curthreads == 0)
			threadpool_cleanup(tpool);
		else
			threadpool_unlock(tpool);
	}
}

void
thread_yield(void)
{
	sched_yield();
}

/*
 * XXX hack, uninterruptable (though cancelable) sleep implementation.
 * XXX assumes that an_timespec is same as timespec
 *
 * I'm not convinced thread_sleep is all that useful or I would spend some
 * time doing it right.
 */
void
thread_sleep(an_timespec_t *delay)
{
	int err;
	an_timespec_t wakeup;

	time_future(&wakeup, delay);

	err = pthread_mutex_lock(&sleepmutex);
	assert(err == 0);

	do {
		pthread_cleanup_push((void (*)(void *))pthread_mutex_unlock,
				     &sleepmutex);
		err = pthread_cond_timedwait(&sleepcond, &sleepmutex, &wakeup);
		pthread_cleanup_pop(0);
	} while (err == 0);
	assert(err == EINVAL || err == ETIMEDOUT);

	err = pthread_mutex_unlock(&sleepmutex);
	assert(err == 0);
}

void *
thread_getdata(void)
{
	return pthread_getspecific(nodeos_threaddata);
}

void
thread_setdata(void *data)
{
	int err;

	err = pthread_setspecific(nodeos_threaddata, data);
	assert(err == 0);
}


/*
 * Thread "interrupt" support.
 *
 * With pthreads, signals are the only way we can do this.
 * Pthread cancelation implies the death of the canceled thread,
 * NodeOS interrupts do not.
 */
#include <signal.h>

#define SIG_THREADINTR	SIGUSR1 /* XXX */
sigset_t thread_intrmask;

static void thread_intr_handler(int sig, siginfo_t *si, void *sc);

static void
thread_intr_init(void)
{
	struct sigaction act;

	sigemptyset(&thread_intrmask);
	sigaddset(&thread_intrmask, SIG_THREADINTR);

	/*
	 * Catch relevant signals.
	 */
#ifdef SA_SIGINFO
	act.sa_flags = SA_SIGINFO;
	act.sa_sigaction = thread_intr_handler;
#else
	act.sa_handler = thread_intr_handler;
#endif
	sigemptyset(&act.sa_mask);

	sigaction(SIG_THREADINTR, &act, 0);
	sigaction(SIGILL, &act, 0);
#ifdef SIGEMT
	sigaction(SIGEMT, &act, 0);
#endif
	sigaction(SIGFPE, &act, 0);
	sigaction(SIGBUS, &act, 0);
	sigaction(SIGSEGV, &act, 0);
}

void
thread_interrupt(thread_t *thread)
{
	int err;
	cond_t *cond;

	thread_lock(thread);

	/*
	 * Send a signal to said thread
	 */
	err = pthread_kill(thread->id, SIG_THREADINTR);
	assert(err == 0);

	/*
	 * If thread is in a cond wait, wake it up.
	 * This gets really strange because we must acquire the cond
	 * lock before the thread lock (because of how thread_condwait
	 * works).
	 */
 again:
	cond = thread->cond;
	if (cond) {
		/*
		 * Cond lock is held.  Release the thread lock and wait
		 * for the cond.  Then reaquire the thread lock and see
		 * if it is still waiting on the same cond.  If not,
		 * start over.
		 */
		if (cond_trylock(cond)) {
			thread_unlock(thread);
			cond_lock(cond);
			thread_lock(thread);
			if (thread->cond != cond) {
				cond_unlock(cond);
				goto again;
			}
		}
		icond_broadcast(&cond->cond);
		cond_unlock(cond);
	}

	thread_unlock(thread);
}

void
thread_block_interrupt(void)
{
	thread_current()->flags &= ~THREAD_INTSON;
}

void
thread_unblock_interrupt(void)
{
	thread_current()->flags |= THREAD_INTSON;
}

void
thread_sethandler(thread_handler_t handler)
{
	thread_current()->handler = handler;
}

/*
 * Signal handler for all exceptions and NodeOS thread interrupts
 */
static void
thread_intr_handler(int sig, siginfo_t *si, void *sc)
{
	thread_t *thread = thread_current();
	int intson, type;

	if (thread == 0 || (thread->flags & THREAD_INKERNEL)) {
		os_dump_state(sc);
		panic("NodeOS: got signal %d\n", sig);
	}

	NODEOS_ENTER;

	/*
	 * No handler for thread, just exit
	 */
	 if (thread->handler == 0) {
		pthread_cancel(thread->id);
		return;
	 }

	/*
	 * Otherwise, call the handler with interrupts disabled
	 */
	intson = (thread->flags & THREAD_INTSON);
	thread->flags &= ~THREAD_INTSON;
	if (sig == SIG_THREADINTR) {
		assert(intson);
		type = AN_THREAD_INTR;
	} else
		type = AN_THREAD_TRAP;

	NODEOS_LEAVE;
	thread->handler(type, sc);
	NODEOS_ENTER;

	/* XXX NODEOS_LEAVE sans signal code */
	thread->flags |= intson;
	thread->flags &= ~THREAD_INKERNEL;

	NODEOS_LEAVE;
}

/*
 * Block a thread on the specified condition variable.
 * The thread may be awakened either by the condition being signaled through
 * an API call or because of thread_interrupt.  We do not differentiate the
 * two cases here.
 *
 * Called with thread and condition locked.
 */
void
thread_condwait(cond_t *cond)
{
	thread_t *thread = thread_current();
	assert(thread->cond == 0);

	thread->cond = cond;
	icond_wait(&cond->cond, &cond->lock);
	thread->cond = 0;
}
