/*
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 *
 */
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <nodeos/impl.h>

#include "threadpool.h"
#include "thread.h"
#include "flow.h"
#include "mem.h"

int
threadpool_create(struct flow *flow, struct chan *chan, cpuspec_t *cspec,
		  struct threadpool **out_tpool)
{
	threadpool_t *tpool;
	int err;

	/* Verify the thread params with the underlying implementation */
	err = thread_checkattrs(cspec);
	if (err)
		return err;

	tpool = mem_alloc(sizeof *tpool);
	if (tpool == 0)
		return ENOMEM;

	memset(tpool, 0, sizeof *tpool);
	tpool->flow = flow;
	tpool->chan = chan;
	tpool->params = *cspec;
	ilock_init(&tpool->lock);

	*out_tpool = tpool;
	return 0;
}

void
threadpool_destroy(struct threadpool *tpool)
{
	thread_t *th, *me;

	me = thread_current();
	threadpool_lock(tpool);
	tpool->flags |= TPOOL_DYING;
	for (th = tpool->threads; th != 0; th = th->next)
		if (th != me)
			thread_kill(th);

	if (tpool->curthreads == 0)
		threadpool_cleanup(tpool);
	else
		threadpool_unlock(tpool);
}

/*
 * Similar to threadpool destroy, but kill only threads for a particular
 * channel (and don't destroy the pool!)
 *
 * Channel should be locked.
 */
void
threadpool_kill(struct threadpool *tpool, struct channel *chan)
{
	thread_t *th, *me;

	me = thread_current();
	threadpool_lock(tpool);
	for (th = tpool->threads; th != 0; th = th->next)
		if (th != me && th->chan == chan)
			thread_kill(th);
	threadpool_unlock(tpool);
}

void
threadpool_cleanup(struct threadpool *tpool)
{
	int flags;

	assert(threadpool_islocked(tpool));
	assert(tpool->curthreads == 0);

	flags = tpool->flags;
	threadpool_unlock(tpool);

	if (flags & TPOOL_DYING) {
		ilock_deinit(&tpool->lock);
		mem_free(tpool, sizeof *tpool);
	}
}
